//eprsave.c
//EEPROM saving/loading for IO-Buddy settings
//Bryan E. Topp <betopp@betopp.com> 2018

//This file is part of IO-Buddy. (2018-09-30)

//IO-Buddy is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//IO-Buddy is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with IO-Buddy.  If not, see <https://www.gnu.org/licenses/>.


#include <stdbool.h>
#include "eprsave.h"
#include "iobuddy.h"

#define EPR_ADR_READ 0xA1
#define EPR_ADR_WRITE 0xA0

//Bitbang I2C driver because I _STILL_ can't figure out the damn PIC32 I2C HW
//PB2 = pin 6 = SDA
//PB3 = pin 7 = SCL
//Keep both of them latched at 0, but enable/disable the tristate mode.
//(Don't see any reason to use PIC's actual open-drain mode)

void BBI2CInit()
{
    ANSELB = 0; //Fuck you, this goddamn analog bullshit needs to be off
    CNPUBSET = (1<<2) | (1<<3); //Both need to have internal pull-ups enabled
    TRISBSET = (1<<2) | (1<<3); //Don't drive yet
    LATBCLR = (1<<2) | (1<<3); //When we drive, drive low    
}

//Wait function is called after changing one of the I2C pins, controls rate
void BBI2CWait()
{
    //Tuned with a scope to be a few hundred KHz
    for(volatile int i = 0; i < 5; i++)  {   }
}

void BBI2CDriveSDA()
{
    TRISBCLR = (1<<2);
    BBI2CWait();
}

void BBI2CFloatSDA()
{
    TRISBSET = (1<<2);
    BBI2CWait();
}

void BBI2CDriveSCL()
{
    TRISBCLR = (1<<3);
    BBI2CWait();
}

void BBI2CFloatSCL()
{
    TRISBSET = (1<<3);
    BBI2CWait();
}

bool BBI2CReadSDA()
{
    if(PORTBbits.RB2)
        return true;
    else
        return false;
}

void BBI2CStart()
{
    BBI2CFloatSDA();
    BBI2CFloatSCL();
    BBI2CDriveSDA();
}

void BBI2CStop()
{
    BBI2CDriveSCL();
    BBI2CDriveSDA();
    BBI2CFloatSCL();
    BBI2CFloatSDA();
}

void BBI2CSendAck()
{
    BBI2CDriveSCL();
    BBI2CDriveSDA();
    BBI2CFloatSCL();
}

void BBI2CNoAck()
{
    BBI2CDriveSCL();
    BBI2CFloatSDA();
    BBI2CFloatSCL();
}

void BBI2CWriteByte(uint8_t outbyte)
{
    for(int bit = 7; bit >= 0; bit--)
    {
        BBI2CDriveSCL();

        if(outbyte & (1<<bit))
            BBI2CFloatSDA();
        else
            BBI2CDriveSDA();

        BBI2CFloatSCL();
    }
}

uint8_t BBI2CReadByte()
{
    uint8_t retval = 0;
    
    for(int bit = 7; bit >= 0; bit--)
    {
        BBI2CDriveSCL();
        BBI2CFloatSDA();
        BBI2CFloatSCL();
        
        if(BBI2CReadSDA())
            retval |= (1<<bit);
    }
    
    return retval;
}


void EPRInit()
{
    BBI2CInit();
}

void EPRSave(const uint8_t *in256bytes)
{
    //Write in 16 bursts of 16
    for(int startaddr = 0; startaddr < 256; startaddr += 16)
    {
        BBI2CStart();
        BBI2CWriteByte(EPR_ADR_WRITE);
        BBI2CNoAck();
        BBI2CWriteByte(startaddr);
        BBI2CNoAck();
        for(int byteaddr = startaddr; byteaddr < startaddr + 16; byteaddr++)
        {
            BBI2CWriteByte(in256bytes[byteaddr]);
            BBI2CNoAck();
        }
        
        BBI2CStop();
        
        //Wait for write to complete
        DelayMS(5);
    }    
}



void EPRLoad(uint8_t *out256bytes)
{
    for(int b = 0; b < 256; b++)
    {
        BBI2CStart();
        BBI2CWriteByte(EPR_ADR_WRITE);
        BBI2CNoAck();
        BBI2CWriteByte(b);
        BBI2CNoAck();
        BBI2CNoAck();
        BBI2CStart();
        BBI2CWriteByte(EPR_ADR_READ);
        BBI2CNoAck();
        out256bytes[b] = BBI2CReadByte();
        BBI2CNoAck();
        BBI2CStop();
    }
}