//iobuddy.c
//XInput-to-JVS converter for NAOMI (entry point)
//Bryan E. Topp <betopp@betopp.com> 2018

//This file is part of IO-Buddy. (2018-09-30)

//IO-Buddy is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//IO-Buddy is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with IO-Buddy.  If not, see <https://www.gnu.org/licenses/>.



//This is an application for the PIC32MX220 and MAX485.
//It provides switch inputs to a JVS game from an XInput game pad.

//System consists of three main parts:
//1. JVS I/O bus protocol
//2. USB-to-JVS controls mapping (i.e. which XInput control is which JVS switch bit)
//3. USB host protocol and XInput polling

//We poll the XInput controllers and put their results in a temporary buffer.
//Then the mapping code reads this and fills another buffer with the intended JVS controls.
//When the JVS code decides to respond to a poll from the NAOMI, it reads that buffer.

//External XO at 24MHz because I had them available
#pragma config FNOSC = PRIPLL
#pragma config POSCMOD = EC

//Divide by 3 to get 4MHz at input to PLL
#pragma config FPLLIDIV = DIV_3
//Multiply by 24 to get up to 96MHz clock
#pragma config FPLLMUL = MUL_24
//Divide back down to 48MHz
#pragma config FPLLODIV = DIV_2

//USB uses the same fixed multiplier and output divisor, and we configure the same input divisor
#pragma config UPLLEN = ON
#pragma config UPLLIDIV = DIV_3

//Peripherals also run at full 48MHz
#pragma config FPBDIV = DIV_1

//Show peripheral clock on CLKO (pin 10)
#pragma config OSCIOFNC = ON

//Don't watchdog by default
#pragma config FWDTEN = OFF

//No JTAG - we need those 5v-tolerant pins for the JVS IO
#pragma config JTAGEN = OFF


#include "iobuddy.h"

#include "jvsio.h"
#include "usbhost.h"
#include "mapping.h"
#include "eprsave.h"

//Note to self - gnu89-inline is necessary for this to not explode.
#include <peripheral/system.h>

int main()
{
	SYSTEMConfigPerformance(SYS_FREQ);

    //LED outputs - PA0 and PA1 are always set high.
    //We toggle the tristate setting to turn the LEDs on/off.
    LATASET = 3;
    
    EPRInit(); //Init EEPROM saving/loading
	USBInit(); //Init USB
	MapInit(); //Init control mapping (will trigger load from EEPROM))
	JVSInit(); //Init JVS communication - turns on JVS UART IRQ
	
    //Main loop polls USB at 1000hz and updates an array of the results
	while(1)
	{
		USBUpdate();
		MapUpdate();
	//	JVSUpdate(); //Happens in interrupt context from JVS UART IRQ
	}
	
	//Should never return
	return 0;
}

