//jvsio.c
//JVS I/O bus protocol
//Bryan E. Topp <betopp@betopp.com> 2018

//This file is part of IO-Buddy. (2018-09-30)

//IO-Buddy is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//IO-Buddy is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with IO-Buddy.  If not, see <https://www.gnu.org/licenses/>.

//Special thanks to wiki.pcbotaku.com and TheOnlyJoey's openjvs.

#include "iobuddy.h"
#include "jvsio.h"

#include <peripheral/int.h>
#include <peripheral/uart.h>

#include <string.h>

//Escape character on JVS. If this is received, it's ignored, but the next character is decremented on the wire.
#define JVS_BYTE_ESC 0xD0

//Sync byte on JVS. This marks the start of a new frame.
#define JVS_BYTE_SYNC 0xE0

//Our address in the JVS chain. 0 if we haven't got one (we never host).
static uint8_t JVSLocalAddress = 0;

static uint8_t JVSInTxMode = 0;

static uint8_t JVSTxBuffer[512] = {0}; //Response to send - totally raw, ready for the wire
static uint16_t JVSTxSize = 0; //Total size of Tx Buffer contents
static uint16_t JVSTxNext = 0; //How many bytes we've transmitted so far

//Response that we build as we process commands.
//This is NOT ready for the wire, and is just the payload data
static uint8_t JVSResponseCode = 0;
static uint8_t JVSResponseCommands[512] = {0};
static uint16_t JVSResponseSize = 0;

//If we're receiving, which part of the process are we in
static uint8_t JVSReadState = 0;
enum JVSReadStates
{
	JVS_READ_NONE = 0, //Just finished transmitting, perhaps - waiting for sync.
	JVS_READ_ADDR, //Got a sync byte, awaiting node address RX = FF or our node
	JVS_READ_SIZE, //Got a node address, awaiting payload+cksum size
	JVS_READ_DATA, //Got the sync, address, size. Receiving payload and checksum.
	JVS_READ_DONE //Have a complete packet, need to process the commands
};

static uint8_t JVSRxEscaped = 0; //Whether we've just read an escape character.
static uint8_t JVSRxAddr = 0; //Which node is addressed in the JVS packet we're receiving.
static uint16_t JVSRxSize = 0; //How many bytes we expect in the payload and checksum.
static uint8_t JVSRxBuffer[256] = {0}; //Buffer for payload and checksum bytes.
static uint16_t JVSRxNext = 0; //How many bytes we've received.
static uint8_t JVSRxSum = 0; //Sum of all non-escape bytes following the sync byte.

static uint8_t JVSGotResetCommand = 0; //Whether we're due to reset because of a reset command

//Note the difference between the Tx and Rx buffers above.

#define UARTDEVICE UART2


//This shouldn't be necessary but xc32 seems to somehow break strings that get passed as compiler parms
#define STRINGIZE(x) #x
#define STRINGIZE_VALUE_OF(x) STRINGIZE(x)

//ID that we send in response to IO Board ID command
static const char *JVSID = "Taupe Technologies;IO-Buddy;Rv" STRINGIZE_VALUE_OF(BUILDDATE) ";Configurable USB adapter";

//Player bits that we send in response to the Switch-Input command
uint16_t JVSPlayerBits[4] = {0};

//System bits that we send in response to the Switch Input command
uint8_t JVSSystemBits = 0;

//Analog axis values we send in response to the Analog Input command
uint16_t JVSAnalogValues[4];

//Helper - sets MAX485 to transmit
void JVSSet485TxMode()
{
    JVSInTxMode = 1;
    
	//Pin 16 - output high
    LATBSET = (1<<7); //PORTBbits.RB7 = 1;
    TRISBCLR = (1<<7); //TRISBbits.TRISB7 = 0;
}

//Helper - sets MAX485 to receive
void JVSSet485RxMode()
{
    JVSInTxMode = 0;
    
	//Pin 16 - output low
    LATBCLR = (1<<7); //PORTBbits.RB7 = 0;
    TRISBCLR = (1<<7); //TRISBbits.TRISB7 = 0;
}

//Could be used to implement proper daisy-chaining (this board currently only supports being a sole slave)
int SenseInput()
{
	//Stub - no downstream
	return 2;
}

//Helper - outputs bit on sense line
void SenseOutput(int value)
{
	if(value)
	{
		//Sense high (default) - leave pin 17 alone
        ODCBSET = (1<<8); //ODCBbits.ODCB8 = 1;
        TRISBSET = (1<<8); //TRISBbits.TRISB8 = 1;

        //Also set PA0 - LED - to input, to turn it off
        TRISASET = (1<<0);
	}
	else
	{
		//Sense low (we're done configuring) - pull pin 17 low
        ODCBSET = (1<<8); //ODCBbits.ODCB8 = 1;
        LATBCLR = (1<<8); //PORTBbits.RB8 = 0;
		TRISBCLR = (1<<8); //TRISBbits.TRISB8 = 0;
        
        //Set the LED as an output to turn it on
        TRISACLR = (1<<0); //TRISAbits.TRISA0 = 0;
	}
	
}

//Helper - queues up a response to be sent
void JVSBufferResponseData(uint8_t resp)
{
	JVSResponseCommands[JVSResponseSize] = resp;
	JVSResponseSize++;
}

//Processes a command from part of the RX buffer. Returns number of bytes consumed. May add to Response buffer.
uint16_t JVSProcessCommand(const uint8_t *buffer)
{	
	switch(buffer[0])
	{
		case 0xF0: //Reset
			
		
			if(buffer[1] == 0xD9)
			{
				JVSGotResetCommand = 1;
			}
			
			return 2; //Command input was 2 bytes
		
		case 0xF1: //Set address
		
			if(JVSLocalAddress != 0)
			{
				//We already have an address. Ignore.
                
			}
			else if(SenseInput() == 1)
			{
				//The downstream sense says there's an unconfigured board there.
                //This address is for someone downstream.
			}
			else
			{
				//We don't have an address, nor an unconfigured downstream board.
                //This is our new address.
				JVSLocalAddress = buffer[1];
				
                //Indicate upstream that we've got an address.
				SenseOutput(0);
				
				JVSBufferResponseData(0x01); //Ack
			}
	
			return 2; //Command input 2 bytes
		
		case 0x10: //Get slave ID
		
			
			//Ack and return the board ID
			JVSBufferResponseData(0x01); //Ack
			for(const char *respchar = JVSID; *respchar != '\0'; respchar++)
			{
				JVSBufferResponseData(*respchar); //String
			}
			JVSBufferResponseData(0); //NUL-term
			
			return 1; //Command was 1 byte
		
		case 0x11: //Get command-set revision
		
			JVSBufferResponseData(0x01); //Ack
			JVSBufferResponseData(0x13); //v1.1
			return 1; //Command was 1 byte
		
		case 0x12: //Get JVS revision
		
			JVSBufferResponseData(0x01); //Ack
			JVSBufferResponseData(0x30); //v2.0
			return 1; //Command was 1 byte
		
		case 0x13: //Get communication version

			JVSBufferResponseData(0x01); //Ack
			JVSBufferResponseData(0x10); //v1.1
			return 1;
		
		case 0x14: //Get board functions

			JVSBufferResponseData(0x01); //Ack
		
			JVSBufferResponseData(0x01); //Number of players and switches
			JVSBufferResponseData(4);  //Four players
			JVSBufferResponseData(14); //14 switches (1L8B+ST+SV) per player
			JVSBufferResponseData(0); //Unused parm
		
			JVSBufferResponseData(0x02); //Number of coin slots
			JVSBufferResponseData(2); //Two coin slots
			JVSBufferResponseData(0); //Unused parm
			JVSBufferResponseData(0); //Unused parm
            
            JVSBufferResponseData(0x03); //Number of analog channels and bits
            JVSBufferResponseData(4); //Four channels
            JVSBufferResponseData(8); //Eight bits
            JVSBufferResponseData(0);
            
            JVSBufferResponseData(0x12); //No GPIO
            JVSBufferResponseData(0);
            JVSBufferResponseData(0);
            JVSBufferResponseData(0);
		
			JVSBufferResponseData(0x00); //End
			return 1;
			
		case 0x15: //Transmit board ID

			//printf("ID: [%s]\n", buffer + 1);
			JVSBufferResponseData(0x01); //Ack
			return strlen(buffer+1) + 1;
            
        case 0x16:
            JVSBufferResponseData(0x16);
            return 1;
            
		case 0x20: //Get switch input data
	
            JVSBufferResponseData(0x01); //Ack
            
            JVSBufferResponseData(JVSSystemBits); //System buttons
		
			uint8_t player_index_max = buffer[1];
			uint8_t byte_index_max = buffer[2];
            
            //Per-player buttons
            for(uint8_t player_index = 0; player_index < player_index_max; player_index++)
            {
                for(uint8_t byte_index = 0; byte_index < byte_index_max; byte_index++)
                {
                    //Return control data
                    if(byte_index == 0 && player_index < 4)
                    {
                        //First byte of valid player switches
                        JVSBufferResponseData(JVSPlayerBits[player_index] % 256);
                    }
                    else if(byte_index == 1 && player_index < 4)
                    {
                        //Second byte of valid player switches
                        JVSBufferResponseData(JVSPlayerBits[player_index] / 256);
                    }
                    else
                    {
                        //Byte 0 or invalid player -> return zeros
                        JVSBufferResponseData(0);
                    }
                }
            }
			
			return 3;
            
        case 0x21: //Get coin input
            
            JVSBufferResponseData(0x01); //Ack
            
            int num_slots = buffer[1];
            for(int s = 0; s < num_slots; s++)
            {
                JVSBufferResponseData(0x00); //No coin
                JVSBufferResponseData(0x00); 
            }
            
            return 2;
		
        case 0x22: //Get analog channel
            
            JVSBufferResponseData(0x01); //Ack
            
            int channel_max = buffer[1];
            for(int c = 0; c < channel_max; c++)
            {
                if(c >= 4)
                {
                    //Invalid...
                    JVSBufferResponseData(0x80);
                    JVSBufferResponseData(0x00);
                }
                else
                {
                    JVSBufferResponseData(JVSAnalogValues[c] / 256);
                    JVSBufferResponseData(JVSAnalogValues[c] % 256);
                }
            }
            
            return 2;
            
		default:
			//printf("Didn't recognize command type %2.2X, dropping further commands\n", buffer[0]);
            buffer[0];
            volatile int dbggr = buffer[0];
			JVSResponseCode = 2; //Overall response is a command-unknown
			JVSBufferResponseData(0x02); //Command unknown
			return 255; //Can't re-align after a command we don't recognize - rest of packet is trash		
	}
}

//Processes all commands in already-filled RX buffer.
void JVSProcessRxBuffer()
{
	//It's good!
	
	//We're going to process a number of commands in the packet.
	JVSResponseSize = 0;
	JVSResponseCode = 1; //Normal code, to begin
	
	//Process the commands received.
	uint16_t buf_idx = 0;
	while(buf_idx < JVSRxSize - 1) //last byte is checksum
		buf_idx += JVSProcessCommand(&(JVSRxBuffer[buf_idx]));
    
    JVSRxSize = 0;
    memset(JVSRxBuffer, 0, sizeof(JVSRxBuffer));
}

//Helper for building response - Queues up a sync byte in the transmit buffer.
void JVSTxBufferSyncByte()
{
	JVSTxBuffer[JVSTxSize] = JVS_BYTE_SYNC;
	JVSTxSize++;
}

//Helper for building response - Queues up a normal data byte in the transmit buffer.
//Adds escape character as necessary to avoid sync/escape.
void JVSTxBufferEscapedData(uint8_t dat)
{
	if(dat == JVS_BYTE_SYNC || dat == JVS_BYTE_ESC)
	{
		JVSTxBuffer[JVSTxSize] = JVS_BYTE_ESC;
		JVSTxSize++;
		dat--;
	}
	
	JVSTxBuffer[JVSTxSize] = dat;
	JVSTxSize++;
}

//Builds the raw response data in the Tx Buffer, from the Response buffer
void JVSBuildTxBuffer()
{
	JVSTxSize = 0;
	JVSTxNext = 0;
    
		
    JVSTxBufferSyncByte(); //Sync
	JVSTxBufferEscapedData(0x00); //Address 0 -> responding to host
	uint8_t txsum = 0x00; //Technically should include the address... but that's always 0.
	
	JVSTxBufferEscapedData(1+JVSResponseSize+1); //Size of code + unescaped response + 1 checksum byte
	txsum += 1+JVSResponseSize+1; //Size is included in checksum.
	
	//Response code
	JVSTxBufferEscapedData(JVSResponseCode);
	txsum += JVSResponseCode;
	
	//Response data
	for(uint16_t d = 0; d < JVSResponseSize; d++)
	{
		JVSTxBufferEscapedData(JVSResponseCommands[d]);
		txsum += JVSResponseCommands[d]; //Checksum unescaped data
	}
	
	JVSTxBufferEscapedData(txsum);
}


void JVSUpdateTx()
{	
	//Transmit mode. We're sending a packet that we built.
	JVSSet485TxMode();
    
    DelayUS(100);
    
    //Transmit until we're done - preempting the USB code
    
    while(JVSTxNext < JVSTxSize)
    {
        while(!UARTTransmitterIsReady(UARTDEVICE))
        {      
        }
        
        //When we put together the Tx Buffer, we assemble it ready-to-go, with the
        //escape codes and checksum and everything. So we can just transmit byte-by-byte.

        //Transmit a byte.
        UARTSendDataByte(UARTDEVICE, JVSTxBuffer[JVSTxNext]);
        JVSTxNext++;
    }

    while(!UARTTransmissionHasCompleted(UARTDEVICE))
    {
    
    }
    
    //Drop out of transmit mode, but don't start receiving anything in particular.
    //When we receive a sync byte, we'll initialize the receiver properly.
    JVSTxSize = 0;
    JVSTxNext = 0;
    JVSReadState = JVS_READ_NONE;
        
    JVSSet485RxMode();
}

void JVSUpdateRx()
{	
	//If there's nothing to receive, just abort.
	while(UARTReceivedDataIsAvailable(UARTDEVICE))
	{
        //Grab the byte off the wire and decide what to do.
        uint8_t newbyte = UARTGetDataByte(UARTDEVICE);
        if(newbyte == JVS_BYTE_SYNC)
        {
            //Got a sync byte - reset the read process. Next byte should be node address.
            JVSReadState = JVS_READ_ADDR;
            JVSRxEscaped = 0;
            JVSRxAddr = 0;
            JVSRxSize = 0;
            JVSRxNext = 0;
            JVSRxSum = 0;
        }
        else if(newbyte == JVS_BYTE_ESC)
        {
            //Got an escape character - discard it but set the flag.
            JVSRxEscaped = 1;
        }
        else
        {
            //Not an escape or sync byte. If we were previously escaped, handle that.
            if(JVSRxEscaped)
            {
                newbyte ++;
                JVSRxEscaped = 0;		
            }

            //Process the data.
            switch(JVSReadState)
            {
                case JVS_READ_ADDR:
                    //This was a read-address.
                    JVSRxAddr = newbyte;
                    JVSRxSum += newbyte; //Address is included in checksum.
                    JVSReadState = JVS_READ_SIZE;

                    //If we haven't got an address yet, ignore non-broadcast packets.
                    //If we have an address, also consider packets addressed to us.
                    if(JVSLocalAddress == 0)
                    {
                        //Note that we use "address 0" to mean "we don't have an address"
                        //because this is an implementation for an IO device, not a master.

                        if(JVSRxAddr != 0xFF)
                        {
                            //If we don't have an address, ignore all except broadcasts
                            //JVSReadState = JVS_READ_NONE;
                            
                            //Hack - if address 1 gets a packet, assume that
                            //it's for us and we just rebooted for some reason.
                            //We don't support downstream boards, so we're
                            //always guaranteed to be the first in the chain.
                            if(JVSRxAddr == 0x01)
                            {
                                JVSLocalAddress = 0x01;
                                SenseOutput(0);
                            }
                            else
                            {
                                //Otherwise ignore like we should
                                JVSReadState = JVS_READ_NONE;
                            }
                        }
                    }
                    else
                    {
                        if(JVSRxAddr != 0xFF && JVSRxAddr != JVSLocalAddress)
                        {
                            //If we DO have an address, ignore non-broadcasts to other addresses.
                            JVSReadState = JVS_READ_NONE;
                        }
                    }
                break;

                case JVS_READ_SIZE:
                    //This was a size.
                    JVSRxSize = newbyte;
                    JVSRxSum += newbyte; //Size is included in checksum.
                    JVSReadState = JVS_READ_DATA;
                break;

                case JVS_READ_DATA:
                    //This was a data byte.
                    JVSRxBuffer[JVSRxNext] = newbyte;
                    JVSRxNext++;

                    if(JVSRxNext >= JVSRxSize)
                    {
                        //We got the number of bytes specified in the size.
                        //The packet is done.
                        //The last byte we got should have been the checksum.
                        if(newbyte == JVSRxSum)
                        {
                            //Valid checksum - Kick off the processing and response building.
                            JVSReadState = JVS_READ_DONE;
                            return;
                        }
                        else
                        {
                            //Invalid checksum - just go back to waiting
                            JVSReadState = JVS_READ_NONE;
                        }
                    }
                    else
                    {
                        //This wasn't the last byte.
                        //Keep running the checksum.
                        JVSRxSum += newbyte;
                    }

                break;
            }

        }
    }
}

//Initializes JVS bus I/O
void JVSInit()
{	
	//Start with the sense line pulled to 2.5V by external diodes - do nothing to it ourselves
	SenseOutput(1);
	
	//Start needing an address
	JVSLocalAddress = 0;

	//Start in receive mode
	JVSSet485RxMode();
	
	//Clear read state-machine
	JVSRxSize = 0;
	JVSRxNext = 0;
	JVSRxEscaped = 0;
	JVSReadState = JVS_READ_NONE;
    
    JVSTxSize = 0;
    JVSTxNext = 0;
    
    //Set up UART
    UARTConfigure(UARTDEVICE, UART_ENABLE_PINS_TX_RX_ONLY);
    UARTSetFifoMode(UARTDEVICE, UART_INTERRUPT_ON_RX_NOT_EMPTY);
    UARTSetLineControl(UARTDEVICE, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UARTDEVICE, 48000000UL, 115200UL);
    UARTEnable(UARTDEVICE, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX | UART_TX));

    // Configure UART RX Interrupt
    INTEnable(INT_SOURCE_UART_RX(UARTDEVICE), INT_ENABLED);
    INTSetVectorPriority(INT_VECTOR_UART(UARTDEVICE), INT_PRIORITY_LEVEL_2);
    INTSetVectorSubPriority(INT_VECTOR_UART(UARTDEVICE), INT_SUB_PRIORITY_LEVEL_0);

    // configure for multi-vectored mode
    INTConfigureSystem(INT_SYSTEM_CONFIG_MULT_VECTOR);

    // enable interrupts
    INTEnableInterrupts();
    
    //Set UART pins
    U2RXRbits.U2RXR = 1; //UART 2 RX = Pin B5 = pin 14
    RPB9Rbits.RPB9R = 2; //UART 2 TX = Pin B9 = pin 18
    
    //Turn on the output on the UART TX pin
    TRISBCLR = (1<<9); //TRISBbits.TRISB9 = 0;
    LATBSET = (1<<9); //PORTBbits.RB9 = 1;  
}




//Updates bus communications, processes packets, generates responses
void JVSUpdate()
{	

    
    JVSUpdateRx();
    
	//If we just finished receiving a valid packet, process its commands.
	if(JVSReadState == JVS_READ_DONE)
	{
		JVSProcessRxBuffer();
		
		//If we're not reading anything, but we have responses to send,
		//build the transmit buffer and start transmitting them.
		if(JVSResponseSize > 0 && JVSLocalAddress > 0)
		{
			JVSBuildTxBuffer();
			JVSUpdateTx();
		}
		
		JVSReadState = JVS_READ_NONE;
	}

    if(JVSGotResetCommand)
    {
        JVSLocalAddress = 0;
        JVSGotResetCommand = 0;
    }
    

}

//Sets player input bits that we will respond with next poll.
void JVSSetPlayerBits(uint8_t plidx, uint16_t bitmask)
{
    //We only support 4 players
	if(plidx > 3)
		return;
	
	JVSPlayerBits[plidx] = bitmask;
}

//Sets the system bits for response.
void JVSSetSystemBits(uint8_t bitmask)
{
    JVSSystemBits = bitmask;
}

//Sets an analog axis for response.
void JVSSetAnalogValue(uint8_t axidx, uint16_t value)
{
    if(axidx > 3)
        return;
    
    JVSAnalogValues[axidx] = value;
    
}

//We may be in the middle of a USB transaction when a JVS byte arrives,
//so allow JVS update to run as an interrupt handler too. Guard against reentry.
void __ISR(_UART_2_VECTOR, IPL4) _UART2Handler(void)
{
    //Process receievd character
    JVSUpdate();
    
    IFS1bits.U2RXIF = 0;
} 