//mapping.c
//Pulls controls polled from the pad and stores to JVS data
//Bryan E. Topp <betopp@betopp.com> 2018

//This file is part of IO-Buddy. (2018-09-30)

//IO-Buddy is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//IO-Buddy is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with IO-Buddy.  If not, see <https://www.gnu.org/licenses/>.

#include "mapping.h"
#include "usbhost.h"
#include "jvsio.h"
#include "eprsave.h"

#include <string.h>
#include <stdint.h>
#include <stdbool.h>

//Individual structure detailing a controller mapping
typedef struct controller_map_s
{
    //VID and PID of USB device that this map is for
    uint16_t vid;
    uint16_t pid;
    
    //Mappings:
    //Each JVS button gets three bits here.
    //Those three bits select one of the action buttons on the controller.
    uint32_t buttonsources;
} controller_map_t;

//Data structure that we pack to the EEPROM - alias of the EEPROM shadow
#define MAXSAVEDMAPS 10
typedef struct saved_map_data_s
{
    //How many times the system has been turned on
    uint16_t power_on_count;
    
    //Controller maps we've saved
    controller_map_t controller_maps[MAXSAVEDMAPS];
    
    //Highest number here will get replaced when we make a new map.
    uint8_t map_most_recent[MAXSAVEDMAPS];
    
} saved_map_data_t;

//256-byte array that comes out of the EEPROM
static uint8_t epr_shadow[256];
static saved_map_data_t *savedata = (saved_map_data_t *)epr_shadow;

//Which controller maps are being used for each player
controller_map_t *ActiveMaps[4];

//Remapping status for each player
typedef struct remap_status_s
{    
    //Are they currently in "remapping mode"?
    int IsRemapping;
    
    //How many buttons have they remapped in remapping mode?
    int ButtonsRemappedCount;
    
    //Which actions buttons are already mapped in remapping mode?
    int ButtonsRemappedMask;
    
    //Switch-source settings that we'll commit to the map when they're done
    int NewSourceBits;
    
} remap_status_t;

remap_status_t RemapStatus[4];

void MapInit()
{
    //Get the old EEPROM contents ready, and note the additional power-on
    EPRLoad(epr_shadow);
    savedata->power_on_count++;
    EPRSave(epr_shadow);
    
    //Zero out the remapping state machine
    memset(RemapStatus, 0, sizeof(RemapStatus));
}

static uint16_t MapGetJVSSwitches(int p)
{
    if(p >= 4)
        return 0;
    if(ActiveMaps[p] == NULL)
        return 0;

    uint16_t final_jvs_switches = 0;

    uint32_t actionbuttons = USBGetActionButtons(p);
    uint32_t dirbuttons = USBGetDirectionButtons(p);
    uint32_t sysbuttons = USBGetSystemButtons(p);

    //Handle joystick, start, service - no remapping
    final_jvs_switches |= (dirbuttons & POLLED_DIR_UP) ? JVSPB_UP : 0;
    final_jvs_switches |= (dirbuttons & POLLED_DIR_DOWN) ? JVSPB_DOWN : 0;
    final_jvs_switches |= (dirbuttons & POLLED_DIR_LEFT) ? JVSPB_LEFT : 0;
    final_jvs_switches |= (dirbuttons & POLLED_DIR_RIGHT) ? JVSPB_RIGHT : 0;

    final_jvs_switches |= (sysbuttons & POLLED_SYS_START) ? JVSPB_START : 0;
    
    //Hackish - we use HOME+SELECT for the test switch, so don't also hit SVC
    if(!(sysbuttons & POLLED_SYS_HOME))
        final_jvs_switches |= (sysbuttons & POLLED_SYS_SELECT) ? JVSPB_SVC : 0;
    
    //Handle remappable action buttons.
    const uint16_t jvs_action_buttons[8] = 
    {
        JVSPB_PUSH1,
        JVSPB_PUSH2,
        JVSPB_PUSH3,
        JVSPB_PUSH4,
        JVSPB_PUSH5,
        JVSPB_PUSH6,
        JVSPB_PUSH7,
        JVSPB_PUSH8,
    };

    uint32_t mapdata = ActiveMaps[p]->buttonsources;
    for(int remappable = 0; remappable < 8; remappable++)
    {
        int source_button = mapdata & 0x7;
        if(actionbuttons & (1<<source_button))
            final_jvs_switches |= jvs_action_buttons[remappable];

        mapdata >>= 3;
    }
    
    return final_jvs_switches;
}

//Marks the given controller map as the most-recently-used.
static void MarkMostRecent(int mapidx)
{
    //Shift back the place of anything that was more recently used
    int prior_slot = savedata->map_most_recent[mapidx];
    for(int m = 0; m < MAXSAVEDMAPS; m++)
    {
        if(savedata->map_most_recent[m] < prior_slot)
            savedata->map_most_recent[m]++;
    }
    
    //We're now the most-recent
    savedata->map_most_recent[mapidx] = 0;
}

//Picks a controller mapping for the given player number.
//Stores it in their ActiveMap pointer.
static void PickMap(int p)
{
    //Get VID/PID from the USB host system
    uint16_t currentvid = USBGetPadVID(p);
    uint16_t currentpid = USBGetPadPID(p);
    
    //Look through all mappings we have. If we find a match, use it.
    for(int m = 0; m < MAXSAVEDMAPS; m++)
    {
        if(savedata->controller_maps[m].pid == currentpid)
        {
            if(savedata->controller_maps[m].vid == currentvid)
            {
                //This is the map we want.
                ActiveMaps[p] = &(savedata->controller_maps[m]);
                
                //Mark it as the most-recently-used.
                MarkMostRecent(m);
                
                //Done
                return;
            }
        }
    }
    
    //Okay, we didn't find a mapping...
    //Find the least-recently-used mapping and replace it.
    int least_recent = 0;
    for(int mi = 0; mi < MAXSAVEDMAPS; mi++)
    {
        if(savedata->map_most_recent[mi] > savedata->map_most_recent[least_recent])
            least_recent = mi;
    }
    
    //Take over the VID/PID of this entry
    savedata->controller_maps[least_recent].pid = currentpid;
    savedata->controller_maps[least_recent].vid = currentvid;
    
    //Default the new mapping
    savedata->controller_maps[least_recent].buttonsources = 
            (0 << (0*3)) |
            (1 << (1*3)) |
            (2 << (2*3)) |
            (3 << (3*3)) |
            (4 << (4*3)) |
            (5 << (5*3)) |
            (6 << (6*3)) |
            (7 << (7*3));
 
    //Make it active for this player
    ActiveMaps[p] = &(savedata->controller_maps[least_recent]);
                
    //Mark it as the most-recently-used.
    MarkMostRecent(least_recent);
}

void MapUpdateRemapping(int p)
{
    uint32_t sysbuttons = USBGetSystemButtons(p);
            
    if(!RemapStatus[p].IsRemapping)
    {
        //Not remapping.
        //Allow remapping to start if HOME and START are pressed.
        if((sysbuttons & POLLED_SYS_HOME) && (sysbuttons & POLLED_SYS_START))
        {
            RemapStatus[p].IsRemapping = 1;
            RemapStatus[p].ButtonsRemappedCount = 0;
            RemapStatus[p].ButtonsRemappedMask = 0;
            RemapStatus[p].NewSourceBits = 0;
        }
    }
    else if(sysbuttons & POLLED_SYS_HOME)
    {
        //Currently remapping. Still holding the home button.
        //Look for new buttons presses and track them.
        uint8_t OldActionButtons = RemapStatus[p].ButtonsRemappedMask;
        uint8_t NewActionButtons = USBGetActionButtons(p);

        uint8_t buttons_to_map = NewActionButtons & ~OldActionButtons;
        if(buttons_to_map != 0)
        {
            for(int remap_button = 0; remap_button < 8; remap_button++)
            {
                if(buttons_to_map & (1<<remap_button))
                {
                    //remap_button represents which controller button we've now pressed.
                    //ButtonsRemappedCount represents which JVS switch it will trigger.
                    RemapStatus[p].NewSourceBits |= (remap_button & 0x7) << (3*RemapStatus[p].ButtonsRemappedCount);
                    RemapStatus[p].ButtonsRemappedCount++;
                }
            }
        }

        //Ignore those buttons going forward.
        RemapStatus[p].ButtonsRemappedMask |= buttons_to_map;
    }
    else
    {
        //Remapping, but they released the home button.
        //Finish the remapping process.
        //Remap the unmapped buttons in order
        for(int unmapped_button = 0; unmapped_button < 8; unmapped_button++)
        {
            if( (RemapStatus[p].ButtonsRemappedMask & (1<<unmapped_button)) == 0)
            {
                RemapStatus[p].NewSourceBits |= (unmapped_button & 0x7) << (3*RemapStatus[p].ButtonsRemappedCount);
                RemapStatus[p].ButtonsRemappedCount++;
            }
        }

        //Save the config
        ActiveMaps[p]->buttonsources = RemapStatus[p].NewSourceBits;
        
        EPRSave(epr_shadow); //Might take 100ms or so... should we defer it?
        
        //Clear remap status
        RemapStatus[p].IsRemapping = 0;
    }
}

void MapUpdate()
{
    int anyone_hitting_test_switch = 0;
    
    //Update each player
    for(int p = 0; p < 4; p++)
    {               
        //Get the exact VID and PID so we can pick a map
        uint16_t currentvid = USBGetPadVID(p);
        uint16_t currentpid = USBGetPadPID(p);
        
        if(currentpid == 0 && currentvid == 0)
        {
            //Don't handle players who don't have any pad
            JVSSetPlayerBits(p, 0);
            continue;
        }
        
        //See if the current map needs to change
        if(ActiveMaps[p] == NULL || ActiveMaps[p]->pid != currentpid || ActiveMaps[p]->vid != currentvid)
        {
            //Map is invalid or doesn't match - pick a new one
            PickMap(p);
        }        
        
        //Mapping now matches their pad.
        
        //Handle the controls for remapping buttons
        MapUpdateRemapping(p);
        
        //If they're busy remapping, present all 0s to JVS. Otherwise, do the map.
        if(RemapStatus[p].IsRemapping)
        {
            JVSSetPlayerBits(p, 0);
        }
        else
        {
            JVSSetPlayerBits(p, MapGetJVSSwitches(p));
            
            //Hackish - allow test-switch with this combo
            uint8_t sysb = USBGetSystemButtons(p);
            if( (sysb & POLLED_SYS_HOME) && (sysb & POLLED_SYS_SELECT) )
                anyone_hitting_test_switch++;
        }
    }
    
    if(anyone_hitting_test_switch)
        JVSSetSystemBits(JVSSB_TEST);
    else
        JVSSetSystemBits(0);
    
    //Don't really have any mapping for analog yet...
    JVSSetAnalogValue(0, USBGetAnalogAxis(0, 0));
    JVSSetAnalogValue(1, USBGetAnalogAxis(0, 1));
    JVSSetAnalogValue(2, USBGetAnalogAxis(0, 2));
    JVSSetAnalogValue(3, USBGetAnalogAxis(0, 3));
}