//usbhost.c
//Functions for hosting the USB side of the iobuddy
//Bryan E. Topp <betopp@betopp.com> 2018

//This file is part of IO-Buddy. (2018-09-30)

//IO-Buddy is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//IO-Buddy is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with IO-Buddy.  If not, see <https://www.gnu.org/licenses/>.

//PIC32MX220 configuration bits we care about
#pragma config FUSBIDIO = OFF //Not actually doing OTG detect
#pragma config FVBUSONIO = OFF //Not actually controlling host power

#include "iobuddy.h"
#include "usbhost.h"
#include <string.h>


//Overall state
static int USBInitDone = 0;
static int USBTimeout = 0;
static int USBDeviceLowspeed = 0; //Whether the attached device is low-speed

//Even/odd ping pong state
static int NextRXBD = 0;
static int NextTXBD = 0;

//Address assignments
static uint8_t USBLastAddressAssigned = 0;


//Status that a USB hub can report
#define USB_PORT_STAT_CONNECTION        0x0001
#define USB_PORT_STAT_ENABLE            0x0002
#define USB_PORT_STAT_SUSPEND           0x0004
#define USB_PORT_STAT_OVERCURRENT       0x0008
#define USB_PORT_STAT_RESET             0x0010
#define USB_PORT_STAT_L1                0x0020
/* bits 6 to 7 are reserved */
#define USB_PORT_STAT_POWER             0x0100
#define USB_PORT_STAT_LOW_SPEED         0x0200
#define USB_PORT_STAT_HIGH_SPEED        0x0400
#define USB_PORT_STAT_TEST              0x0800
#define USB_PORT_STAT_INDICATOR         0x1000

//Features we can set/clear on a hub
#define HUB_FEATURE_C_HUB_LOCAL_POWER           0
#define HUB_FEATURE_C_HUB_OVER_CURRENT          1
#define HUB_FEATURE_PORT_CONNECTION             0
#define HUB_FEATURE_PORT_ENABLE                 1
#define HUB_FEATURE_PORT_SUSPEND                2
#define HUB_FEATURE_PORT_OVER_CURRENT           3
#define HUB_FEATURE_PORT_RESET                  4
#define HUB_FEATURE_PORT_POWER                  8
#define HUB_FEATURE_PORT_LOW_SPEED              9
#define HUB_FEATURE_C_PORT_CONNECTION           16
#define HUB_FEATURE_C_PORT_ENABLE               17
#define HUB_FEATURE_C_PORT_SUSPEND              18
#define HUB_FEATURE_C_PORT_OVER_CURRENT         19
#define HUB_FEATURE_C_PORT_RESET                20
#define HUB_FEATURE_PORT_TEST                   21
#define HUB_FEATURE_PORT_INDICATOR              22

//8-byte "setup" packet we send for USB control transfers
typedef struct usb_setup_packet_s
{
    uint8_t bmRequestType;
    uint8_t bRequest;
    uint16_t wValue;
    uint16_t wIndex;
    uint16_t wLength;    
} usb_setup_packet_t;

#define BM_REQTYPE_DIR_OUT (0 << 7)
#define BM_REQTYPE_DIR_IN (1 << 7)

#define BM_REQTYPE_TYPE_STANDARD (0 << 5)
#define BM_REQTYPE_TYPE_CLASS (1 << 5)
#define BM_REQTYPE_TYPE_VENDOR (2 << 5)

#define BM_REQTYPE_RECIPIENT_DEVICE (0 << 0)
#define BM_REQTYPE_RECIPIENT_INTERFACE (1 << 0)
#define BM_REQTYPE_RECIPIENT_ENDPOINT (2 << 0)
#define BM_REQTYPE_RECIPIENT_OTHER (3 << 0)

#define REQUEST_GET_STATUS 0x00
#define REQUEST_CLEAR_FEATURE 0x01
#define REQUEST_SET_FEATURE 0x03
#define REQUEST_SET_ADDRESS 0x05
#define REQUEST_GET_DESCRIPTOR 0x06
#define REQUEST_SET_DESCRIPTOR 0x07
#define REQUEST_GET_CONFIGURATION 0x08
#define REQUEST_SET_CONFIGURATION 0x09



//When we enumerate devices, we look for up to 4 gamepads to poll.

//Enum of each type of gamepad we can support
enum PadType_e
{
    PAD_NONE = 0, //Saw something we didn't recognize
    PAD_XINPUT, //Matches Xbox 360 controller
    PAD_PS3, //Matches Playstation 3 controller
    PAD_PS4, //Matches Playstation 4 controller
    PAD_MAX
};

//Information about one gamepad that we've detected on the USB bus
typedef struct PadInfo_s
{
    uint8_t PadType; //What kind of gamepad this is
    uint8_t USBAddress; //What's the device's USB address
    uint8_t USBEndpointIn; //Which endpoint to read
    uint8_t USBEndpointOut; //Which endpoint to write
    
    uint16_t USBVID; //VID of their device
    uint16_t USBPID; //PID of their device
    
    //Data we stored at last poll
    uint8_t PolledActionButtons;
    uint8_t PolledDirectionButtons;
    uint8_t PolledSystemButtons;
    
    //Analog data (not well-implemented...)
    //JVS wants it unsigned, 16-bit
    uint16_t AnalogAxes[4];
} PadInfo_t;

//All the info about the USB devices we care about
static PadInfo_t PadInfos[4];
static uint8_t NumPadInfos = 0;

//Buffer descriptor as we write it out to the hardware
typedef struct BufferDescriptorWrite_s
{
    uint32_t dummy_a : 2;
    uint32_t bstall : 1;
    uint32_t dts : 1;
    uint32_t ninc : 1;
    uint32_t keep : 1;
    uint32_t data01 : 1;
    uint32_t uown : 1;
    uint32_t dummy_b : 8;
    uint32_t byte_count : 10;
    uint32_t dummy_c : 6;
    
    uint32_t buffer_address : 32;
    
} BufferDescriptorWrite_t;

//Buffer descriptor as we read back from the hardware
typedef struct BufferDescriptorRead_s
{
    uint32_t dummy_a : 2;
    uint32_t pid : 4;
    uint32_t data01 : 1;
    uint32_t uown : 1;
    uint32_t dummy_b : 8;
    uint32_t byte_count : 10;
    uint32_t dummy_c : 6;
    
    uint32_t buffer_address : 32;
    
} BufferDescriptorRead_t;

//BDT - buffer descriptor table
static BufferDescriptorWrite_t BufferDescriptors[4]  __attribute__((aligned(512)));
enum BufferDescriptor_e
{
    BD_RX_EVEN = 0,
    BD_RX_ODD = 1,
    BD_TX_EVEN = 2,
    BD_TX_ODD = 3
};

uint32_t PhysAddr(void *addr)
{
    uint32_t virt = (unsigned) addr;
    uint32_t phys;

    if (virt & 0x80000000) {
        if (virt & 0x40000000) {
            phys = virt;
        } else {
            phys = virt & 0x1fffffff;
        }
    } else {
        phys = virt + 0x40000000;
    }
    return phys;
}

void WaitIdle()
{
    int tries = 0;
    while(U1CONbits.TOKBUSY) {
        DelayUS(1);
        tries++;
        if(tries > 1024)
            break;
    }
    
    U1IRbits.TRNIF = 1;
    NextTXBD = 0;
    NextRXBD = 0;
}

BufferDescriptorWrite_t *GetNextTXDescriptor()
{
    if(NextTXBD == 0)
        return &(BufferDescriptors[BD_TX_EVEN]);
    else
        return &(BufferDescriptors[BD_TX_ODD]);
}

BufferDescriptorWrite_t *GetNextRXDescriptor()
{
    if(NextRXBD == 0)
        return &(BufferDescriptors[BD_RX_EVEN]);
    else
        return &(BufferDescriptors[BD_RX_ODD]);
}

void USBSetup(usb_setup_packet_t *pkt, uint8_t endpt)
{
    BufferDescriptorWrite_t *bd = GetNextTXDescriptor();
    memset(bd, 0, sizeof(*bd));

    bd->buffer_address = PhysAddr(pkt);
    bd->byte_count = 8;
    bd->uown = 1;

    U1TOK = 0xD0 | (endpt & 0xF); //SETUP to EP
    WaitIdle();
}

void USBOut(uint8_t *data, uint8_t len, uint8_t endpt)
{
    BufferDescriptorWrite_t *bd = GetNextTXDescriptor();
    memset(bd, 0, sizeof(*bd));

    uint8_t d01flag = 0; //First 0/1 flag is 0
    
    bd->buffer_address = PhysAddr(data);
    bd->byte_count = len;
    bd->data01 = d01flag;
    bd->uown = 1;
    

    U1TOK = 0x10 | (endpt & 0xF); //OUT to EP
    WaitIdle();  
    
    while( (bd->byte_count > 0) && (bd->byte_count < len) )
    {
        len -= bd->byte_count;
        data += bd->byte_count;
        
        bd = GetNextTXDescriptor();
        memset(bd, 0, sizeof(*bd));

        bd->buffer_address = PhysAddr(data);
        bd->byte_count = len;
        d01flag ^= 1;
        bd->data01 = d01flag; //Alternate flags
        bd->uown = 1;

        U1TOK = 0x10 | (endpt & 0xF); //OUT to EP
        WaitIdle();        
    }
}

void USBOutNoData(uint8_t endpt)
{
    BufferDescriptorWrite_t *bd = GetNextTXDescriptor();
    memset(bd, 0, sizeof(*bd));

    uint8_t d01flag = 0; //First 0/1 flag is 0
    
    static uint8_t dummy_data = 0;
    bd->buffer_address = PhysAddr(&dummy_data);
    bd->byte_count = 0;
    bd->data01 = d01flag;
    bd->uown = 1;

    U1TOK = 0x10 | (endpt & 0xF); //OUT to EP
    WaitIdle();
}

uint8_t USBIn(uint8_t *data, uint8_t len, uint8_t endpt)
{
    BufferDescriptorWrite_t *bd = GetNextRXDescriptor();
    memset(bd, 0, sizeof(*bd));
    
    //uint8_t totalread = 0;
    uint8_t d01flag = 1; //First 0/1 flag is 1
    
    bd->buffer_address = PhysAddr(data);
    bd->byte_count = len;
    bd->data01 = d01flag;
    bd->uown = 1;
    
    U1TOK = 0x90 | (endpt & 0xF); //IN to EP
    WaitIdle();
    //totalread += bd->byte_count;
    
    //continue getting data
    
    while( (bd->byte_count > 0) && (bd->byte_count < len) )
    {
        len -= bd->byte_count;
        data += bd->byte_count;
        
        bd = GetNextRXDescriptor();
        memset(bd, 0, sizeof(*bd));

        bd->buffer_address = PhysAddr(data);
        bd->byte_count = len;
        d01flag ^= 1;
        bd->data01 = d01flag; //Alternate flags
        bd->uown = 1;

        U1TOK = 0x90 | (endpt & 0xF); //IN to EP
        WaitIdle();
        //totalread += bd->byte_count;
    }
    
    const BufferDescriptorRead_t *bd_read = (BufferDescriptorRead_t*)bd;
    if(bd_read->pid != 0)
        USBTimeout = 0;
    
    return bd_read->pid; //Return handshake
}

void USBInNoData(uint8_t endpt)
{
    BufferDescriptorWrite_t *bd = GetNextRXDescriptor();
    memset(bd, 0, sizeof(*bd));
    
    uint8_t d01flag = 1; //First 0/1 flag is 1
    
    static uint8_t dummy_data = 0;
    bd->buffer_address = PhysAddr(&dummy_data);
    bd->byte_count = 0;
    bd->data01 = d01flag;
    bd->uown = 1;
    
    U1TOK = 0x90 | (endpt & 0xF); //IN to EP
    WaitIdle();

    const BufferDescriptorRead_t *bd_read = (BufferDescriptorRead_t*)bd;
    if(bd_read->pid != 0)
        USBTimeout = 0;
}

//I'm not really handling NAKs from the device... so just retry the command
#define USB_DESC_RETRY_COUNT 10
#define USB_DESC_RETRY_WAIT_MS 25

//Gets descriptor from USB device. otbuf must have enough room for it.
void USBGetDescriptor(uint8_t addr, uint8_t desctype, uint8_t descindex, uint8_t *otbuf, uint16_t len)
{
    //It doesn't make sense to call this while disregarding the output...
    if(len <= 0)
        return;
    if(otbuf == NULL)
        return;
    
    //Zero the first byte of the output buffer.
    //Later, check if it's nonzero (indicating nonzero size, i.e. a descriptor)
    otbuf[0] = 0;
    
    //Make attempts to get the descriptor.
    for(uint8_t attempt = 0; attempt < USB_DESC_RETRY_COUNT; attempt++)
    {
        WaitIdle();

        //Send get_descriptor command
        usb_setup_packet_t pkt;
        pkt.bmRequestType = BM_REQTYPE_DIR_IN | BM_REQTYPE_TYPE_STANDARD | BM_REQTYPE_RECIPIENT_DEVICE;
        pkt.bRequest = REQUEST_GET_DESCRIPTOR;
        pkt.wValue = (descindex | (desctype << 8));
        pkt.wIndex = 0;
        pkt.wLength = len;

        U1ADDRbits.DEVADDR = addr;

        USBSetup(&pkt, 0);

        //Get data back
        USBIn(otbuf, len, 0);

        //Acknowledge it
        USBOutNoData(0);
        
        //If we got a valid packet, stop retrying.
        //If we didn't get a valid response, the USB DMA leave this alone...
        //Meaning we'd still see the zero we put there initially.
        if(otbuf[0] != 0)
            break;
        
        //Otherwise, delay and try again
        DelayMS(USB_DESC_RETRY_WAIT_MS);
    }
}

//Gets descriptor from USB class. otbuf must have enough room for it.
void USBGetClassDescriptor(uint8_t addr, uint8_t desctype, uint8_t descindex, uint8_t *otbuf, uint16_t len)
{
    //It doesn't make sense to call this while disregarding the output...
    if(len <= 0)
        return;
    if(otbuf == NULL)
        return;

    otbuf[0] = 0;
    for(uint8_t attempt = 0; attempt < USB_DESC_RETRY_COUNT; attempt++)
    {
        WaitIdle();

        usb_setup_packet_t pkt;
        pkt.bmRequestType = BM_REQTYPE_DIR_IN | BM_REQTYPE_TYPE_CLASS | BM_REQTYPE_RECIPIENT_DEVICE;
        pkt.bRequest = REQUEST_GET_DESCRIPTOR;
        pkt.wValue = (descindex | (desctype << 8));
        pkt.wIndex = 0;
        pkt.wLength = len;

        U1ADDRbits.DEVADDR = addr;

        USBSetup(&pkt, 0);
        USBIn(otbuf, len, 0);
        USBOutNoData(0);

        if(otbuf[0] != 0)
            break;

        DelayMS(USB_DESC_RETRY_WAIT_MS);
    }
}

void USBSetConfig(uint8_t addr, uint8_t newconfig)
{
    //Send config assignment command
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_OUT | BM_REQTYPE_TYPE_STANDARD | BM_REQTYPE_RECIPIENT_DEVICE;
    pkt.bRequest = REQUEST_SET_CONFIGURATION;
    pkt.wValue = newconfig;
    pkt.wIndex = 0;
    pkt.wLength = 0;
    
    U1ADDRbits.DEVADDR = addr;
    
    USBSetup(&pkt, 0);
    
    //Zero-length data back
    USBInNoData(0);
}

//Assigns the current 0-addressed device to the next free address.
uint8_t USBAssignAddress()
{
    uint8_t newaddr = USBLastAddressAssigned + 1;
    
    //Send address assignment command
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_OUT | BM_REQTYPE_TYPE_STANDARD | BM_REQTYPE_RECIPIENT_DEVICE;
    pkt.bRequest = REQUEST_SET_ADDRESS;
    pkt.wValue = newaddr;
    pkt.wIndex = 0;
    pkt.wLength = 0;

    U1ADDRbits.DEVADDR = 0; //If we're assigning an address, it's to a device currently on 0
    
    USBSetup(&pkt, 0);

    //Zero-length data back
    USBInNoData(0);
    
    USBLastAddressAssigned = newaddr;
    return newaddr;
}

//Sets a feature flag on a port of a USB hub.
void USBSetPortFeature(uint8_t hubaddr, uint8_t hubport, uint8_t feature)
{
    WaitIdle();

    //Send set_port_feature
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_OUT | BM_REQTYPE_TYPE_CLASS | BM_REQTYPE_RECIPIENT_OTHER;
    pkt.bRequest = REQUEST_SET_FEATURE;
    pkt.wValue = feature;
    pkt.wIndex = hubport;
    pkt.wLength = 0;

    U1ADDRbits.DEVADDR = hubaddr;

    USBSetup(&pkt, 0);
    USBInNoData(0);
}

void USBClearPortFeature(uint8_t hubaddr, uint8_t hubport, uint8_t feature)
{
    WaitIdle();

    //Send set_port_feature
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_OUT | BM_REQTYPE_TYPE_CLASS | BM_REQTYPE_RECIPIENT_OTHER;
    pkt.bRequest = REQUEST_CLEAR_FEATURE;
    pkt.wValue = feature;
    pkt.wIndex = hubport;
    pkt.wLength = 0;

    U1ADDRbits.DEVADDR = hubaddr;

    USBSetup(&pkt, 0);
    USBInNoData(0);
}

//Sets a feature flag on a device.
void USBSetDevFeature(uint8_t addr, uint8_t feature)
{
    WaitIdle();

    //Send set_port_feature
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_OUT | BM_REQTYPE_TYPE_STANDARD | BM_REQTYPE_RECIPIENT_DEVICE;
    pkt.bRequest = REQUEST_SET_FEATURE;
    pkt.wValue = feature;
    pkt.wIndex = 0;
    pkt.wLength = 0;

    U1ADDRbits.DEVADDR = addr;

    USBSetup(&pkt, 0);
    USBInNoData(0);
}

//Gets the status of a USB port on a hub.
uint32_t USBGetPortStatus(uint8_t hubaddr, uint8_t hubport)
{
    WaitIdle();

    //Send get_port_status
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_IN | BM_REQTYPE_TYPE_CLASS | BM_REQTYPE_RECIPIENT_OTHER;
    pkt.bRequest = REQUEST_GET_STATUS;
    pkt.wValue = 0;
    pkt.wIndex = hubport;
    pkt.wLength = 4;

    U1ADDRbits.DEVADDR = hubaddr;

    USBSetup(&pkt, 0);

    //Get data back
    uint32_t retval = 0;
    USBIn(&retval, 4, 0);

    //Acknowledge it
    USBOutNoData(0);
    
    return retval;
}

//Gets status of a device
uint16_t USBGetDevStatus(uint8_t addr)
{
    WaitIdle();

    //Send get_status
    usb_setup_packet_t pkt;
    pkt.bmRequestType = BM_REQTYPE_DIR_IN | BM_REQTYPE_TYPE_STANDARD | BM_REQTYPE_RECIPIENT_DEVICE;
    pkt.bRequest = REQUEST_GET_STATUS;
    pkt.wValue = 0;
    pkt.wIndex = 0;
    pkt.wLength = 2;

    U1ADDRbits.DEVADDR = addr;

    USBSetup(&pkt, 0);

    //Get data back
    uint16_t retval = 0;
    USBIn(&retval, 2, 0);

    //Acknowledge it
    USBOutNoData(0);
    
    return retval;
    
}


//Calls USBAssignAddress, then checks if the new device is a hub, and 
//then recurses and assigns addresses to each port
void USBAssignAddressRecursive()
{
    //Query the new device's device descriptor to see if it's a hub
    uint8_t devdesc[12] = {0};
    USBGetDescriptor(0 /*addr*/, 0x01, 0, devdesc, 12);
    
    //Assign an address to the device on address 0.
    //This may be something on a hub, or on our root port.
    uint8_t newaddr = USBAssignAddress();
    

    uint8_t devclass = devdesc[4];
    uint8_t devsubclass = devdesc[5];
    uint8_t devprotocol = devdesc[6];
    if(devclass == 9 && devsubclass == 0 && devprotocol == 0)
    {
        //This is a hub.
        USBSetConfig(newaddr, 0x01); //Set config 01
        DelayMS(10);
 
        //Get hub descriptor to find out how many ports it has
        uint8_t hubdesc[16] = {0};
        USBGetClassDescriptor(newaddr, 0x29, 0, hubdesc, 16);
        DelayMS(10);
        uint8_t ports = hubdesc[2];
        
        //Turn on power
        for(int hubport = 1; hubport <= ports; hubport++)
        {
            USBSetPortFeature(newaddr, hubport, HUB_FEATURE_PORT_POWER);
            DelayMS(10);
        }
        
        //Bring its ports out of reset, in order, and try to assign addresses.
        for(int hubport = 1; hubport <= ports; hubport++)
        {
            //Turn on power to port
            USBSetPortFeature(newaddr, hubport, HUB_FEATURE_PORT_ENABLE);
            DelayMS(10);
            
            
            //Get port status, see if it's reporting a connection change
            uint32_t portstat = USBGetPortStatus(newaddr, hubport);
            if(portstat & (USB_PORT_STAT_CONNECTION<<16))
            {                
                //Something's connected.
                //Clear the just-connected flag
                USBClearPortFeature(newaddr, hubport, HUB_FEATURE_C_PORT_CONNECTION);

                //Wait 100ms
                DelayMS(100);
                //Reset port
                USBSetPortFeature(newaddr, hubport, HUB_FEATURE_PORT_RESET);
                DelayMS(10);
                
                //Wait until reset-change
                while(USBGetPortStatus(newaddr, hubport) & USB_PORT_STAT_RESET)
                {
                    DelayMS(10);
                }

                //Here it would be proper to check if the port reported
                //a reset change, I guess... but some hubs don't?
                DelayMS(10);
                //portstat = USBGetPortStatus(newaddr, hubport);
                if(1) //if(portstat & (USB_PORT_STAT_RESET << 16))
                {
                    //Clear just-reset flag
                    DelayMS(10);
                    USBClearPortFeature(newaddr, hubport, HUB_FEATURE_C_PORT_RESET);
                    DelayMS(10);
                    
                    portstat = USBGetPortStatus(newaddr, hubport);
                    if(portstat & USB_PORT_STAT_ENABLE)
                    {
                        //Assign an address to the new device
                        USBAssignAddressRecursive();
                    }
                }
                
                
            }    
        }
        
        
        
    }
}

//Finds input and output endpoint descriptors following an interface descriptor
void USBFindEndpointsFollowingInterface(const uint8_t *if_descriptor, uint8_t *result_ep_in, uint8_t *result_ep_out)
{
    int descriptors_skipped = 0;
    const uint8_t *descptr = if_descriptor;
    while(descptr[1] != 0x05 && descriptors_skipped++ < 10)
        descptr += descptr[0]; //Skip descriptors until we come to an endpoint
    
    uint8_t first_endpoint = descptr[2];
    
    descptr += descptr[0];
    while(descptr[1] != 0x05 && descriptors_skipped++ < 10)
        descptr += descptr[0]; //Skip descriptors until we come to an endpoint
    
    uint8_t second_endpoint = descptr[2];
   
    //Check if we got valid endpoint descriptors - one in and one out
    uint8_t ep_in = 0x81;
    uint8_t ep_out = 0x02;
    if( (first_endpoint & 0x80) && !(second_endpoint & 0x80) )
    {
        ep_in = first_endpoint;
        ep_out = second_endpoint;
    }
    else if( (second_endpoint & 0x80) && !(first_endpoint & 0x80) )
    {
        ep_out = first_endpoint;
        ep_in = second_endpoint;
    }
    
    *result_ep_in = ep_in;
    *result_ep_out = ep_out;
}


//Given a configured USB address, prepares it to be polled as a pad
void USBInitXboxPad(uint8_t padindex, uint8_t usbaddr, uint8_t *if_descriptor)
{    
    //Read past the interface descriptor to find first two endpoint descriptors
    uint8_t ep_in = 0;
    uint8_t ep_out = 0;
    USBFindEndpointsFollowingInterface(if_descriptor, &ep_in, &ep_out);
    
    //Store the player-number-to-usb-addr mapping
    PadInfos[padindex].USBAddress = usbaddr;
    PadInfos[padindex].USBEndpointIn = ep_in;
    PadInfos[padindex].USBEndpointOut = ep_out;
    PadInfos[padindex].PadType = PAD_XINPUT;
    
    //Set lights
    uint8_t data[8];
    data[0] = 0x01;
    data[1] = 0x03;
    data[2] = 2 + padindex;
    USBOut(data, 3, PadInfos[padindex].USBEndpointOut);    

}

void USBPollXboxPad(uint8_t padindex)
{
    U1ADDRbits.DEVADDR = PadInfos[padindex].USBAddress;
    
    uint8_t inp[16] = {0};
    USBIn(inp, 16, PadInfos[padindex].USBEndpointIn);

    //Responses starting with 0x00, 0x14 are pad data
    if(inp[0] == 0x00 && inp[1] == 0x14)
    {
        //Bytes 2 through 19 make up the data...
        
        //Byte 3 is almost all action buttons,
        //so use it outright (but zero the home-button and unused flags))
        PadInfos[padindex].PolledActionButtons = inp[3] & 0xF3;
        
        //Check if we should report an LT/RT press - include in action buttons
        if(inp[4] > 64) //LT
            PadInfos[padindex].PolledActionButtons |= 0x04;
        if(inp[5] > 64) //RT
            PadInfos[padindex].PolledActionButtons |= 0x08;
        
        //Byte 2, low 4 contain directions on D-pad, same as we define them
        PadInfos[padindex].PolledDirectionButtons = inp[2] & 0x0F;
        
        //"System buttons" are start, select, and home
        PadInfos[padindex].PolledSystemButtons = 0;
        if(inp[2] & 0x10)
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_START;
        if(inp[2] & 0x20)
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_SELECT;
        if(inp[3] & 0x04)
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_HOME;
        
        //Report first analog axis = right X stick
        int32_t signed_rx = ((int16_t)inp[10]) + ((int16_t)inp[11] << 8);
        
        //Xbox controller reports signed, we want unsigned.
        PadInfos[padindex].AnalogAxes[0] = (signed_rx + 0x8000);
        
        //Report second analog axis as right trigger, which we get unsigned 8-bit
        PadInfos[padindex].AnalogAxes[1] = inp[5] << 8;
        
        //Then left trigger
        PadInfos[padindex].AnalogAxes[2] = inp[4] << 8;
        
        //Report last analog axis as dummy for now...
        PadInfos[padindex].AnalogAxes[3] = 0x8000;
    }    
}

void USBInitPS4Pad(uint8_t padindex, uint8_t usbaddr, uint8_t *if_descriptor)
{
    //Read past the interface descriptor to find first two endpoint descriptors
    uint8_t ep_in = 0;
    uint8_t ep_out = 0;
    USBFindEndpointsFollowingInterface(if_descriptor, &ep_in, &ep_out);
    
    //Store the player-number-to-usb-addr mapping
    PadInfos[padindex].USBAddress = usbaddr;
    PadInfos[padindex].USBEndpointIn = ep_in;
    PadInfos[padindex].USBEndpointOut = ep_out;
    PadInfos[padindex].PadType = PAD_PS4;
    
    //Todo - set the LED color
}

void USBPollPS4Pad(uint8_t padindex)
{
    U1ADDRbits.DEVADDR = PadInfos[padindex].USBAddress;
    
    uint8_t inp[16] = {0};
    USBIn(inp, 16, PadInfos[padindex].USBEndpointIn);
    
    //PS4 HID report format starts with 0x01
    if(inp[0] = 0x01)
    {
        //Byte 5 high bits contain face buttons
        PadInfos[padindex].PolledActionButtons = inp[5] & 0xF0;
        
        //Byte 6 low bits contain triggers
        PadInfos[padindex].PolledActionButtons |= inp[6] & 0x0F;
        
        //Do system buttons...
        PadInfos[padindex].PolledSystemButtons = 0;
        if(inp[6] & 0x20) //Button "Option" = bit 5 of byte 6
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_START;
        if(inp[6] & 0x10) //Button "share" = bit 4 of byte 6
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_SELECT;
        if(inp[7] & 0x01) //Button "PS" = bit 0 of byte 7
            PadInfos[padindex].PolledSystemButtons |= POLLED_SYS_HOME;
        
        //D-pad reports like a hat...
        uint8_t HatToDirButtons[16] = 
        {
            POLLED_DIR_UP,
            POLLED_DIR_UP | POLLED_DIR_RIGHT,
            POLLED_DIR_RIGHT,
            POLLED_DIR_RIGHT | POLLED_DIR_DOWN,
            POLLED_DIR_DOWN,
            POLLED_DIR_DOWN | POLLED_DIR_LEFT,
            POLLED_DIR_LEFT,
            POLLED_DIR_LEFT | POLLED_DIR_UP,
            0, 0, 0, 0, 0, 0, 0, 0
        };
        PadInfos[padindex].PolledDirectionButtons = HatToDirButtons[inp[5] & 0x0F];
        
        //Temp analog axes...
        PadInfos[padindex].AnalogAxes[0] = inp[3] << 8; //RX
        PadInfos[padindex].AnalogAxes[1] = inp[9] << 8; //R2
        PadInfos[padindex].AnalogAxes[2] = inp[8] << 8; //L2
        PadInfos[padindex].AnalogAxes[3] = 0x8000;
    }
    
}

void USBInit()
{    
    //Asserts, effectively
    while(sizeof(BufferDescriptorWrite_t) != 8) { };
    while(sizeof(BufferDescriptorRead_t) != 8) { };
    while((uint32_t)BufferDescriptors & 0x000001FF != 0) { };
    
    USBInitDone = 0;

    //Turn off LED to indicate no devices yet
    TRISASET = (1<<1); //TRISAbits.TRISA1 = 1;
    
    //Toggle USB module off, then on
    DelayMS(10);
    U1PWRCbits.USBPWR = 0;
    DelayMS(10);
    U1PWRCbits.USBPWR = 1;
    DelayMS(10);
    
    //Enable interrupts from it
    IEC1bits.USBIE = 1;
       
    //Configure SOF timing
    U1SOF = 75;
    
    //Configure endpoints
    U1EP0bits.EPRXEN = 1;
    U1EP0bits.EPTXEN = 1;
    U1EP0bits.EPHSHK = 1;
    U1EP0bits.RETRYDIS = 0; //Don't disable retries during setup
    
    //Configure buffer descriptor table
    memset(BufferDescriptors, 0, sizeof(BufferDescriptors));
    
    U1CONbits.PPBRST = 1; //Fuck it, single-buffer
    
    //Set location of buffer descriptor table for hardware
    U1BDTP3 = (PhysAddr(BufferDescriptors) >> 24) & 0xFF;
    U1BDTP2 = (PhysAddr(BufferDescriptors) >> 16) & 0xFF;
    U1BDTP1 = (PhysAddr(BufferDescriptors) >> 8) & 0xFE;
    
    U1CONbits.SOFEN = 0; //To reduce noise on the bus, disable the SOF packet generation by writing the SOF enable bit to ?0?, the SOFEN bit (U1CON<0>) = 0.
    U1IEbits.ATTACHIE = 1; //Enable the device attach interrupt, the ATTACHIE bit (U1IE<6> = 1).
    U1IEbits.UERRIE = 1;
    U1IEbits.TRNIE = 1;
    U1IRbits.ATTACHIF = 1;
    U1CONbits.HOSTEN = 1; //Enable Host mode, the HOSTEN bit (U1CON<3>) = 1.

    //Wait for attachment. Clear interrupt flag after it happens.
    //Hack - sometimes we miss an attach?
    for(int waiter = 0; waiter < 100; waiter++)
    {
        if(U1IRbits.ATTACHIF)
        {
            U1IRbits.ATTACHIF = 1;
            break;
        }
        DelayMS(1);
    }
    
    //Sleep 100ms after attachment for power to stabilize.
    DelayMS(100);
    
    //Check full-speed or low-speed
    if(U1CONbits.JSTATE == 0 && 0 /*HACK: seem to spuriously detect low-speed*/)
    {
        //low-speed
        U1ADDRbits.LSPDEN = 1; 
        U1EP0bits.LSPD = 1;
        USBDeviceLowspeed = 1;
    }
    else
    {
        //full-speed
        U1ADDRbits.LSPDEN = 0; 
        U1EP0bits.LSPD = 0;
        USBDeviceLowspeed = 0;
    }

    //Start reset
    U1CONbits.USBRST = 1;

    //Wait 50ms in reset.
    DelayMS(50);
    
    //End reset
    U1CONbits.USBRST = 0;
    
    DelayMS(10);
    
    //Start SOF generation
    U1CONbits.SOFEN = 1;
    
    //Wait another 50ms after reset before actually driving traffic
    DelayMS(50);

    //Assign an address to the connected device, and recurse through hubs
    USBLastAddressAssigned = 0; //So first device is 1
    USBAssignAddressRecursive();
    
    //Give device some time after assigning address
    DelayMS(50);
    
    //Iterate through all attached devices, looking for pads    
    NumPadInfos = 0;
    for(uint8_t addr = 1; addr <= USBLastAddressAssigned; addr++)
    {        
        //Examine the device at address "addr".
        //Only consider the first configuration (do any controllers use others?)
        uint8_t configidx = 0; 
        
        //Get the device descriptor, to record the VID/PID
        static uint8_t descbuffer[1024];
        uint16_t dev_vid = 0;
        uint16_t dev_pid = 0;
        
        USBGetDescriptor(addr, 0x01, 0, descbuffer, 12);
        DelayMS(50);
        
        dev_vid = descbuffer[8] + (descbuffer[9] << 8);
        dev_pid = descbuffer[10] + (descbuffer[11] << 8);
        
        
        //Get the complete descriptor hierarchy in two parts...
        
        //Get the first 9 bytes, which should contain the length of it
        USBGetDescriptor(addr, 0x02, configidx, descbuffer, 9);
        DelayMS(50);
        
        //Get the rest
        uint16_t desclen = descbuffer[2] + (descbuffer[3] << 8);
        
        USBGetDescriptor(addr, 0x02, configidx, descbuffer, desclen);
        DelayMS(50);
        
        //Set device to the first configuration
        uint8_t config_selection = descbuffer[5];
        USBSetConfig(addr, config_selection);   
        DelayMS(50);
        
        //Move through the descriptors returned looking at each interface.

        //Each descriptor starts with a byte indicating its length.
        //So if we're looking at a descriptor starting at "offs", the byte
        //at that offset will tell us where the next descriptor starts.
        for(uint16_t offs = 0; offs < desclen; offs += descbuffer[offs])
        {
            uint8_t *descptr = &(descbuffer[offs]);     
            if(descptr[0] == 0)
            {
                //Bad descriptor - zero length. We'll never finish.
                break;
            }
            
            if(descptr[1] == 0x04)
            {
                //This is an interface descriptor.                
                uint8_t interface_number = descptr[2];
                uint8_t class = descptr[5];
                uint8_t subclass = descptr[6];
                uint8_t protocol = descptr[7];
                
                if(class == 255 && subclass == 93 && protocol == 1)
                {
                    //Found an xbox 360 controller interface.
                    
                    //Record device IDs
                    PadInfos[NumPadInfos].USBPID = dev_pid;
                    PadInfos[NumPadInfos].USBVID = dev_vid;
                    
                    //Init pad endpoints
                    USBInitXboxPad(NumPadInfos, addr, descptr);
                    
                    //Done
                    NumPadInfos++;
                    break;
                }
                
                if(class == 3 && subclass == 0 && protocol == 0)
                {
                    //Found a USB HID device...
                    //Identify based on size of HID report descriptor
                    const uint8_t *hid_desc_ptr = descptr;
                    while(hid_desc_ptr[1] != 0x21)
                        hid_desc_ptr += hid_desc_ptr[0];
                    
                    uint16_t hid_report_desc_size = hid_desc_ptr[7] + (hid_desc_ptr[8] << 8);
                    if(hid_report_desc_size == 0x1F3)
                    {
                        //This is a PS4 controller

                        PadInfos[NumPadInfos].USBPID = dev_pid;
                        PadInfos[NumPadInfos].USBVID = dev_vid;

                        //Init pad endpoints
                        USBInitPS4Pad(NumPadInfos, addr, descptr);

                        //Done
                        NumPadInfos++;
                        break;
                    }
                }
            }
        }
        
        //Only care about the first 4 things we find
        if(NumPadInfos >= 4)
            break;
        
    }
    
    //Success.
    USBInitDone = 1;
    
    if(NumPadInfos > 0)
    {
        //Indicate USB device success by turning LED on
        TRISACLR = (1<<1); //TRISAbits.TRISA1 = 0;
    }
    
    USBTimeout = 0;
    
    //When polling normally, just skip over bad data
    U1EP0bits.RETRYDIS = 1;
}

void USBUpdate()
{       
    //Check timeout
    USBTimeout++;
    if(USBTimeout > 250)
    {
        USBInit();
        return;
    }
    
    //Wait for a new USB frame to start
    uint8_t oldframe = U1FRMLbits.FRML;
    while(U1FRMLbits.FRML == oldframe) { }
    
    //Poll each pad in turn
    for(int p = 0; p < NumPadInfos; p++)
    {
        switch(PadInfos[p].PadType)
        {
            case PAD_XINPUT:
                USBPollXboxPad(p);
            break;
            
            case PAD_PS4:
                USBPollPS4Pad(p);
            break;
            
            default:
            break;
        }
    }  
}

uint16_t USBGetPadVID(uint8_t player)
{
    if(player >= NumPadInfos)
        return 0;
    
    return PadInfos[player].USBVID;
}

uint16_t USBGetPadPID(uint8_t player)
{
    if(player >= NumPadInfos)
        return 0;
    
    return PadInfos[player].USBPID;
}

uint8_t USBGetActionButtons(uint8_t player)
{
    if(player >= NumPadInfos)
        return 0;
    
    return PadInfos[player].PolledActionButtons;
}

uint8_t USBGetDirectionButtons(uint8_t player)
{
    if(player >= NumPadInfos)
        return 0;
    
    return PadInfos[player].PolledDirectionButtons;    
}

uint8_t USBGetSystemButtons(uint8_t player)
{
    if(player >= NumPadInfos)
        return 0;
    
    return PadInfos[player].PolledSystemButtons;    
}

uint16_t USBGetAnalogAxis(uint8_t player, uint8_t axis)
{
    if(player >= NumPadInfos)
        return 0x8000;
    
    if(axis >= 4)
        return 0x8000;
    
    return PadInfos[player].AnalogAxes[axis];
}